import os
import numpy as np
import glob
import time
import operator
import datetime

import iris
import iris.cube
import iris.analysis
import iris.coords
import iris.coord_categorisation
import iris.analysis
from iris.experimental.equalise_cubes import equalise_attributes
from iris.time import PartialDateTime

import warnings
import numba
from numba import jit
from numba import vectorize
from numba import guvectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver


def compute_siarea(cube_gphit, cube_e1t, cube_e2t, cube_sic, masks):
    start = datetime.datetime.now()
    gphit, e1t, e2t, sic_slices = load_data(cube_gphit, cube_e1t, cube_e2t,
                                            cube_sic)
    extn, exts, arean, areas, device = check_device_and_compute(gphit, e1t,
                                                                e2t,
                                                                sic_slices,
                                                                masks)
    seconds_total = datetime.datetime.now() - start
    print('Total ellapsed time on the {1}: {0}'.format(seconds_total, device))
    return extn, exts, arean, areas


def check_device_and_compute(gphit, e1t, e2t, sic_slices, basins):
    check_gpu = numba.cuda.is_available()
    if check_gpu is False:
        device = 'CPU'
        extn, exts, arean, areas = compute_sic_cpu(gphit, e1t, e2t, sic_slices,
                                                   basins)
    else:
        use_gpu = True
        if use_gpu is True:
            device = 'GPU'
            extn, exts, arean, areas = compute_sic_gpu(gphit, e1t, e2t,
                                                       sic_slices, basins)
        else:
            device = 'CPU'
            extn, exts, arean, areas = compute_sic_cpu(gphit, e1t, e2t,
                                                       sic_slices, basins)

    return extn, exts, arean, areas, device


def compute_sic_cpu(gphit, e1t, e2t, sic_slices, basins):
    extn = {}
    exts = {}
    arean = {}
    areas = {}

    for sic_data in sic_slices:
        time = sic_data.coord('time').points[0]
        extn[time] = {}
        exts[time] = {}
        arean[time] = {}
        areas[time] = {}
        sic = sic_data.data
        for basin, tmask in basins.items():
            temp = extn_numba(gphit, e1t, e2t, tmask, sic)
            aux = sum_red.reduce(temp, axis=1)
            extn[time][basin] = sum_red.reduce(aux, axis=1)

            temp = exts_numba(gphit, e1t, e2t, tmask, sic)
            aux = sum_red.reduce(temp, axis=1)
            exts[time][basin] = sum_red.reduce(aux, axis=1)

            temp = arean_numba(gphit, e1t, e2t, tmask, sic)
            aux = sum_red.reduce(temp, axis=1)
            arean[time][basin] = sum_red.reduce(aux, axis=1)

            temp = areas_numba(gphit, e1t, e2t, tmask, sic)
            aux = sum_red.reduce(temp, axis=1)
            areas[time][basin] = sum_red.reduce(aux, axis=1)

    del gphit, e1t, e2t, sic_slices, temp, aux

    return extn, exts, arean, areas


def compute_sic_gpu(gphit, e1t, e2t, sic_slices, basins):
    extn = {}
    exts = {}
    arean = {}
    areas = {}

    gpu_gphit = cuda.to_device(gphit.astype(np.float32))
    gpu_e1t = cuda.to_device(e1t.astype(np.float32))
    gpu_e2t = cuda.to_device(e2t.astype(np.float32))
    gpu_temp = cuda.device_array(e1t.shape[1]*e1t.shape[2], dtype=np.float32)

    block = (128, 1)
    grid_size = ((e1t.shape[2]) // block[0]) + 1
    grid = (grid_size, e1t.shape[1])

    for sic_data in sic_slices:
        time = sic_data.coord('time').points[0]
        extn[time] = {}
        exts[time] = {}
        arean[time] = {}
        areas[time] = {}
        sic = sic_data.data.astype(np.float32)
        gpu_sic = cuda.to_device(sic.astype(np.float32))
        for basin, tmask in basins.items():
            gpu_tmask = cuda.to_device(tmask.astype(np.float32))

            extn_numba_cuda[grid, block](gpu_gphit, gpu_e1t, gpu_e2t,
                                         gpu_tmask, gpu_sic, gpu_temp)
            extn[time][basin] = sum_red_cuda(gpu_temp)

            exts_numba_cuda[grid, block](gpu_gphit, gpu_e1t, gpu_e2t,
                                         gpu_tmask, gpu_sic, gpu_temp)
            exts[time][basin] = sum_red_cuda(gpu_temp)

            arean_numba_cuda[grid, block](gpu_gphit, gpu_e1t, gpu_e2t,
                                          gpu_tmask, gpu_sic, gpu_temp)
            arean[time][basin] = sum_red_cuda(gpu_temp)

            areas_numba_cuda[grid, block](gpu_gphit, gpu_e1t, gpu_e2t,
                                          gpu_tmask, gpu_sic, gpu_temp)
            areas[time][basin] = sum_red_cuda(gpu_temp)

    del gphit, e1t, e2t, sic_slices
    del gpu_gphit, gpu_e1t, gpu_e2t, gpu_tmask, gpu_sic, gpu_temp

    return extn, exts, arean, areas


def load_data(cube_gphit, cube_e1t, cube_e2t, cube_sic):
    gphit = cube_gphit.data.astype(np.float32)
    e1t = cube_e1t.data.astype(np.float32)
    e2t = cube_e2t.data.astype(np.float32)
    sic_slices = []
    for sic_data in cube_sic.slices_over('time'):
        sic_data.data = np.ma.filled(sic_data.data, 0.0).astype(np.float32)
        sic_slices.append(sic_data)

    return gphit, e1t, e2t, sic_slices


def load_cube(filepath, variable):
    cubes = iris.load(filepath, variable)
    if len(cubes) > 1:
        equalise_attributes(cubes)
        iris.util.unify_time_units(cubes)
        cube = cubes.concatenate_cube()
    else:
        cube = cubes[0]
    return cube


@vectorize(['float32(float32, float32, float32, float32, float32)'],
           target='cpu')
def arean_numba(gphit, e1t, e2t, tmask, sic):
    if gphit > 0:
        temp = e1t * e2t * tmask * sic / 100
    return temp


@cuda.jit()
def arean_numba_cuda(gphit, e1t, e2t, tmask, sic, temp):
    i, j = cuda.grid(2)
    x = i + j * e1t.shape[2]
    if(i >= e1t.shape[2]):
        return
    temp[x] = 0.0
    if(gphit[0, j, i] > 0):
        temp[x] = (e1t[0, j, i] * e2t[0, j, i] * tmask[0, j, i] * sic[j, i]
                   / 100)


@vectorize(['float32(float32, float32, float32, float32, float32)'],
           target='cpu')
def extn_numba(gphit, e1t, e2t, tmask, sic):
    if(gphit > 0 and sic > 15):
        temp = e1t * e2t * tmask
    return temp


@cuda.jit()
def extn_numba_cuda(gphit, e1t, e2t, tmask, sic, temp):
    i, j = cuda.grid(2)
    x = i + j * e1t.shape[2]
    if(i >= e1t.shape[2]):
        return
    temp[x] = 0.0
    if(gphit[0, j, i] > 0 and sic[j, i] > 15):
        temp[x] = e1t[0, j, i] * e2t[0, j, i] * tmask[0, j, i]


@vectorize(['float32(float32, float32, float32, float32, float32)'],
           target='cpu')
def areas_numba(gphit, e1t, e2t, tmask, sic):
    if gphit < 0:
        temp = e1t * e2t * tmask * sic / 100
    return temp


@cuda.jit()
def areas_numba_cuda(gphit, e1t, e2t, tmask, sic, temp):
    i, j = cuda.grid(2)
    x = i + j * e1t.shape[2]
    if(i >= e1t.shape[2]):
        return
    temp[x] = 0.0
    if(gphit[0, j, i] < 0):
        temp[x] = (e1t[0, j, i] * e2t[0, j, i] * tmask[0, j, i] * sic[j, i]
                   / 100)


@vectorize(['float32(float32, float32, float32, float32, float32)'],
           target='cpu')
def exts_numba(gphit, e1t, e2t, tmask, sic):
    if(gphit < 0 and sic > 15):
        temp = e1t * e2t * tmask
    return temp


@cuda.jit()
def exts_numba_cuda(gphit, e1t, e2t, tmask, sic, temp):
    i, j = cuda.grid(2)
    x = i + j * e1t.shape[2]
    if(i >= e1t.shape[2]):
        return
    temp[x] = 0.0
    if(gphit[0, j, i] < 0 and sic[j, i] > 15):
        temp[x] = e1t[0, j, i] * e2t[0, j, i] * tmask[0, j, i]


@vectorize(['float64(float32, float32)', 'float64(float64, float64)'],
           target='cpu')
def sum_red(x, y):
    return x+y


@cuda.reduce
def sum_red_cuda(x, y):
    return x+y
